// --------------------------------------------------------------------------
// Copyright 1998-2000 Andrey Radchenko. All Rights Reserved.
// --------------------------------------------------------------------------
// Base elements of the HPGL/2 file. Unit HPGL_2 uses these elements to
// convert HPGL/2 data into proprietary information, useable for displaying
// and printing of the HPGL/2 file.

// --------------------------------------------------------------------------
unit HP_Elem;
interface
uses
  Classes, Windows, Graphics, Controls, ContNrs, Messages, Forms,
  Dialogs;

const
  HPGLSteps = 1016;                // Steps per inch for the HPGL/2
  RTLSteps  = 720;                 // Steps per inch for the RTL jobs
  RtlToHp : real = 1016.0 / 720.0;
  {}
type
  THpFillType = (ftSolid, ftLines, ftCross, ftShade, ftBitmap);
  THpFillInfo = record
                  Color   : TColor;        // Fill color
                  Pattern : THpFillType;   // Fill style
                  Data    : integer;       // Angle or shade of the fill color
                end;
  {}
  TTextInfoType = record
                    Point : TPoint;
                    Angle,
                    Height : integer;
                    Data   : String;
                  end;
  PTextInfoType = ^TTextInfoType;
  {}
  TImageBuffer = array[0..0] of byte;
  PImageBuffer = ^TImageBuffer;
  {}
  TTempNodes = array of TPoint;
  PTempNodes = ^TTempNodes;
  {}
  TImageData   = record
                   Width,                     // Image size
                   Height,
                   Resolut,                   // Image resolution
                   Bits,                      // How many bits per picsel
                   Bytes,                     // Image size in bytes
                   X, Y    : integer;         // Image location
                   Image   : PImageBuffer;    // Image data
                 end;
  TImageDataPtr = ^TImageData;
  {}
  THpNodeList = class(TList)
  public
    destructor  Destroy; override;
    {}
    procedure   Add(Pt : TPoint);
    function    Get(Index : integer) : TPoint;
    procedure   Clear; override;
  end;
  PHpNodeList = ^THpNodeList;
  PPoint    = ^TPoint;
(* Graphical elements of the HPGL/2 list :
   Polyline, polygon, arc, pie slice (filled arc), window (clipping region),
   text string and an image. Circle does not use nodes at all, only
   bounding box. For arc they store two points at arc ends.
   The same applies to a pie.
*)
  TElemType = (etNone, etPline, etPgon, etCircle, etArc, etPie,
               etWin, etFullWin, etText, etImage);

  TElementList = class;
  {}
  TCustomElem = class(TObject)
  private
    fType      : TElemType;
    fTempNodes,                   // Temporary nodes for drawing
    fNodes     : array of TPoint; // Element nodes in the world coords
    fBounds    : TRect;           // Element's bounding box
    fOutline   : TColor;          // Outline color of the element
    fOutStyle  : TPenStyle;       // Outline line style
    fFillStyle : THpFillInfo;
    fCanvas    : TCanvas;         // Drawing canvas
    fCount     : integer;         // Nodes count
    fRWidth    : real;            // Pen width in mm
    fMerge     : boolean;         // How vectors are drawn.
    fParent    : TElementList;
    fTranspar  : boolean;         //
    fBmW,
    fBmH       : integer;         // Dimensions of the temp bitmap
    {}
    procedure SetCanvas(const Value: TCanvas);
    procedure SetFill(const Value: TColor);
    procedure SetOutline(const Value: TColor);
    procedure SetMerge(const Value: boolean);
    procedure SetRealWidth(const Value: real);
    procedure SetPattern(const Value: THpFillType);
    procedure SetCanvasParams;
    procedure ScaleNodes;
    procedure DrawTransparent(Bmp : hBitmap; X, Y : integer; BClr : TColor);
    {}
  public
    constructor Create; virtual;
    destructor  Destroy; override;
    procedure   Draw; virtual;
    procedure   AddPoints(Src : THpNodeList); overload; virtual;
    procedure   AddPoints(Src : PTextInfoType); overload; virtual;
    procedure   AddPoints(Src : TImageDataPtr); overload; virtual;
    procedure   Assign(var Dst : TCustomElem);
    {}
    property  ECanvas : TCanvas read fCanvas write SetCanvas;
    property  ElementType : TElemType read fType;
    property  Bounds : TRect read fBounds;
    property  Fill    : TColor read fFillStyle.Color write SetFill;
    property  Merge   : boolean read fMerge write SetMerge;
    property  Outline : TColor read fOutline write SetOutline;
    property  Pattern : THpFillType read fFillStyle.Pattern write SetPattern;
    property  RealWidth : real read fRWidth write SetRealWidth;
    property  Parent : TElementList read fParent write fParent;
  end;
  {}
  TCustomElemPtr = ^TCustomElem;
  {----------------------------------------------------------}
  TCustomElemArray = array[0..MaxListSize] of TCustomElemPtr;
  PCustomElemArray = ^TCustomElemArray;
  {List of pointers to the HPGL elements}
  {----------------------------------------------------------}
  TElementList = class
  private
    fList       : TObjectList;      // Array of pointers to HP Elements
    fCount      : integer;          // Number of the elements in list
    fRealBounds,                    // Bounding box of the elements, excl. win
    fListBounds : TRect;            // Bounding box of entire plot
    fCanvas     : TCanvas;          // Drawing canvas
    fWidth      : real;             // Current pen width
    fBkColor,                       // Background color
    fOutline    : TColor;           // Outline colors
    fFill       : THpFillInfo;      // Current fill style
    fLineStyle  : TPenStyle;        // Current line style
    fViewPort,                      // Coordinates of the screen viewport
    fView       : TRect;            // Coordinates of the current view
    fResources,                     // Out of resources flag
    fTranspar,                      // Transparency mode
    fUseTranspar,                   // Toggle use of the transparency mode
    fMerge,                         // Merge control
    fDoBounds,
    fDrawing,                       // Flag indicating that list being drawn
    fAborted,                       // User abort flag
    fViewOk,                        // Flag indicating no errors
    fViewReady  : boolean;          // Idicates that list can be drawn
    fClip       : hRgn;             // Handle of the current clip region
    fCanvasRes,
    fXShift,                        // Vertical and horizontal shifts
    fYShift     : integer;          // for screen output.
    fXScale,
    fYScale     : real;             // Scaling factor from world to screen
    {}
    procedure SetCanvas(const Value: TCanvas);
    procedure SetView(const Value: TRect);
    procedure SetViewReady(const Value: boolean);
    procedure SetShiftAndScale;
    function  Compare(Val1, Val2, Op : integer) : integer;
    function  InRange(Elem : TCustomElem) : boolean;
    procedure SetDefaults;
    function  RealWidth(Wd: real): integer;
    procedure SetViewPort(const Value : TRect);
    {}
  public
    constructor Create;
    destructor  Destroy; override;
    function    GetType(const Index : integer) : TElemType;
    procedure   InitView(const ViewPort : TRect);
    function    DevToWorld(const DPoint : TPoint) : TPoint;
    function    WorldToDev(const WPoint : TPoint) : TPoint;
    procedure   SetViewFromScreen(const Value : TRect);
    {}
    procedure Clear;
    function  Add(eType : TElemType; Nodes : Pointer) : boolean;
    property  Aborted : boolean read fAborted write fAborted;
    property  BackColor : TColor read fBkColor write fBkColor;
    function  DrawList : integer;
    property  Count : integer read fCount;
    property  Canvas : TCanvas read fCanvas write SetCanvas;
    property  CurrFill : THpFillInfo read fFill write fFill;
    property  CurrPen  : TColor read fOutline write fOutline;
    property  PenStyle : TPenStyle read fLineStyle write fLineStyle;
    property  PenWidth  : real read fWidth write fWidth;
    property  View : TRect read fView write SetView;
    property  ViewPort : TRect read fViewPort write SetViewPort;
    property  ViewReady : boolean read fViewReady write SetViewReady;
    property  ListBounds : TRect read fListBounds;
    property  RealBounds : TRect read fRealBounds;
    property  XScale : real read fXScale;
    property  YScale : real read fYScale;
    property  XShift : integer read fXShift;
    property  YShift : integer read fYShift;
    property  ClipArea : hRgn read fClip write fClip;
    property  Drawing : boolean read fDrawing;
    property  Merge   : boolean read fMerge write fMerge;
    property  TransMode : boolean read fTranspar write fTranspar;
    property  UseTransMode : boolean read fUseTranspar write fUseTranspar;
  end;
  {----------------------------------------------------------}
  TPolyline = class(TCustomElem)
  public
    constructor Create; override;
    destructor  Destroy; override;
    procedure   Draw; override;
    procedure   AddPoints(Src : THpNodeList); override;
  end;
  {}
  TPolygon = class(TCustomElem)
  public
    constructor Create; override;
    destructor  Destroy; override;
    procedure   Draw; override;
    procedure   AddPoints(Src : THpNodeList); override;
  end;
  {}
  TArc = class(TCustomElem)
  public
    constructor Create; override;
    destructor  Destroy; override;
    procedure Draw; override;
    procedure AddPoints(Src : THpNodeList); override;
  end;
  {}
  TCircle = class(TCustomElem)
  public
    constructor Create; override;
    destructor  Destroy; override;
    procedure Draw; override;
    procedure AddPoints(Src : THpNodeList); override;
  end;
  {}
  TWindow = class(TCustomElem)
  public
    constructor Create; override;
    destructor  Destroy; override;
    procedure Draw; override;
    procedure AddPoints(Src : THpNodeList); override;
  end;
  {}
  TFullWindow = class(TCustomElem)
  public
    constructor Create; override;
    destructor  Destroy; override;
    procedure Draw; override;
  end;
  {}
  TPie = class(TCustomElem)
  public
    constructor Create; override;
    destructor  Destroy; override;
    procedure Draw; override;
    procedure AddPoints(Src : THpNodeList); override;
  end;
  {}
  TText = class(TCustomElem)
  private
    fPoint  : TPoint;   // Text insertion point;
    fText   : string;   // Actual text string;
    fAngle  : integer;  // Angle of the label
    fHeight : integer;  // Height of the lable text;
  public
    constructor Create; override;
    destructor  Destroy; override;
    procedure   Draw; override;
    procedure   AddPoints(Src : PTextInfoType); reintroduce; override;
  end;
  {}
  THpImage = class(TCustomElem)
  private
    fWidth,
    fHeight,
    fDPI   : integer;
    fImage : PImageBuffer;
  public
    constructor Create; override;
    destructor  Destroy; override;
    procedure   Draw; override;
    procedure   AddPoints(Src : TImageDataPtr); reintroduce; override;
  end;
  {}
  function ShadeOfColor(Clr : TColor; Shade : integer) : TColor;
  {}
const
  ElemNames : array[TElemType] of string[8] = ('None', 'Polyline', 'Polygon',
                                               'Circle', 'Arc', 'Pie',
                                               'Window', 'NoWindow','Text',
                                               'Image');
implementation
uses
  SysUtils;
{------------------------------------------------------}
function ShadeOfColor(Clr : TColor; Shade : integer) : TColor;
var
  R, G, B, S : integer;
begin
  S := Round(2.55 * (100 - Shade));
  R := GetRValue(Clr);
  G := GetGValue(Clr);
  B := GetBValue(Clr);
  R := R + S;
  G := G + S;
  B := B + S;
  if R > 255 then
    R := 255;
  if G > 255 then
    G := 255;
  if B > 255 then
    B := 255;
  Result := RGB(R, G, B);
end;
{-----------------------------------------------------}
{ THpNodeList }
procedure THpNodeList.Add(Pt: TPoint);
var
  P : PPoint;
begin
  try
    New(P);
  except
    Exit;
  end;
  P^ := Pt;
  inherited Add(P);
end;
{-----------------------------------------------------}
function THpNodeList.Get(Index: integer) : TPoint;
begin
  if Index >= Count then
    raise ERangeError.Create('Node index out of bounds');
  Result := TPoint(Items[Index]^);
end;
{------------------------------------------------------}
procedure THpNodeList.Clear;
var
  I : integer;
begin
  for I := 0 to Count - 1 do
    Dispose(Items[I]);
  inherited Clear;
end;
{------------------------------------------------------}
destructor THpNodeList.Destroy;
var
  I : integer;
begin
  for I := 0 to Count - 1 do
    Dispose(Items[I]);
  inherited;
end;
{------------------------------------------------------}
constructor TCustomElem.Create;
begin
  inherited Create;
  fBounds.Top := -1 * (MaxInt - 1);
  fBounds.Bottom := MaxInt;
  fBounds.Left := MaxInt;
  fBounds.Right := -1 * (MaxInt - 1);
  fCanvas := nil;
  fType := etNone;
  fOutline := clBlack;
  fFillStyle.Color := clNone;
  fFillStyle.Pattern := ftSolid;
  fRWidth := 0.35;
  fMerge := False;
  fNodes := nil;
  fTempNodes := nil;
  fCount := -1;
end;
{------------------------------------------------------}
destructor TCustomElem.Destroy;
begin
  if Assigned(fNodes) then
    fNodes := nil;
  inherited;
end;
{---------------------------------------------------}
// Prototype for all elements, except text and image
procedure TCustomElem.AddPoints(Src: THpNodeList);
begin
end;
{---------------------------------------------------}
// Prototype for a text element
procedure TCustomElem.AddPoints(Src: PTextInfoType);
begin
end;
{---------------------------------------------------}
// Prototype for an image element
procedure TCustomElem.AddPoints(Src: TImageDataPtr);
begin
end;
{------------------------------------------------------}
procedure TCustomElem.ScaleNodes;
var
  I : integer;
begin
  if fCount < 2 then
    Exit;
  try
    SetLength(fTempNodes, fCount);
  except
    fTempNodes := nil;
    Exit;
  end;
  for I := 0 to fCount - 1 do
    begin
      fTempNodes[I].X := Round(fNodes[I].X * fParent.XScale + fParent.XShift);
      fTempNodes[I].Y := Round(fNodes[I].Y * fParent.YScale + fParent.YShift);
    end;
end;
{---------------------------------------------------}
// Element drawing with merge on. "Borrowed" from the Ms Knowledge base.
// At present time this procedure raises exception in a zoom mode due to
// the size of bitmaps, it creates. Do not know yet how to fix this problem.
procedure TCustomElem.DrawTransparent(Bmp : hBitmap; X, Y : integer;
                                      BClr : TColor);
var
//  Bm : BITMAP;
  Clr : TColor;
  bmAndBack, bmAndObj, bmAndMem, bmSave    : hBitmap;
  bmBackOld, bmObjOld, bmMemOld, bmSaveOld : hBitmap;
  dcMem, dcBack, dcObj, dcTemp, dcSave     : hDC;
  ptSize : TPoint;
  OldMap : integer;
begin
  try
    dcTemp := CreateCompatibleDC(fCanvas.Handle);
    SelectObject(dcTemp, Bmp);
    ptSize.X := fBmW;
    ptSize.Y := fBmH;
    DPtoLP(dcTemp, ptSize, 1);
    fBmW := ptSize.X;
    fBmH := ptSize.Y;
    dcBack := CreateCompatibleDC(fCanvas.Handle);
    dcObj := CreateCompatibleDC(fCanvas.Handle);
    dcMem := CreateCompatibleDC(fCanvas.Handle);
    dcSave := CreateCompatibleDC(fCanvas.Handle);
    {}
    bmAndBack := CreateBitmap(fBmW, fBmH, 1, 1, nil);
    bmAndObj := CreateBitmap(fBmW, fBmH, 1, 1, nil);
    bmAndMem := CreateCompatibleBitmap(fCanvas.Handle, fBmW, fBmH);
    bmSave := CreateCompatibleBitmap(fCanvas.Handle, fBmW, fBmH);
    {}
    bmBackOld := SelectObject(dcBack, bmAndBack);
    bmObjOld := SelectObject(dcObj, bmAndObj);
    bmMemOld := SelectObject(dcMem, bmAndMem);
    bmSaveOld := SelectObject(dcSave, bmSave);
    {}
    OldMap := SetMapMode(fCanvas.Handle, MM_TEXT);
//    SetMapMode(dcTemp, GetMapMode(fCanvas.Handle));
    {}
    BitBlt(dcSave, 0, 0, fBmW, fBmH, dcTemp, 0, 0, SrcCopy);
    Clr := SetBkColor(dcTemp, BClr); //clWhite);
    BitBlt(dcObj, 0, 0, fBmW, fBmH, dcTemp, 0, 0, SrcCopy);
    SetBkColor(dcTemp, Clr);
    BitBlt(dcBack, 0, 0, fBmW, fBmH, dcObj, 0, 0, NotSrcCopy);
    BitBlt(dcMem, 0, 0, fBmW, fBmH, fCanvas.Handle, X, Y, SrcCopy);
    BitBlt(dcMem, 0, 0, fBmW, fBmH, dcObj, 0, 0, SrcAnd);
    BitBlt(dcTemp, 0, 0, fBmW, fBmH, dcBack, 0, 0, SrcAnd);
    BitBlt(dcMem, 0, 0, fBmW, fBmH, dcTemp, 0, 0, SrcPaint);
    BitBlt(fCanvas.Handle, X, Y, fBmW, fBmH, dcMem, 0, 0, SrcAnd);
    BitBlt(dcTemp, 0, 0, fBmW, fBmH, dcSave, 0, 0, SrcCopy);
  finally
    DeleteObject(SelectObject(dcBack, bmBackOld));
    DeleteObject(SelectObject(dcObj, bmObjOld));
    DeleteObject(SelectObject(dcMem, bmMemOld));
    DeleteObject(SelectObject(dcSave, bmSaveOld));
    {}
    DeleteDC(dcMem);
    DeleteDC(dcBack);
    DeleteDC(dcObj);
    DeleteDC(dcSave);
    DeleteDC(dcTemp);
    SetMapMode(fCanvas.Handle, OldMap);
  end;
end;
{------------------------------------------------------}
procedure TCustomElem.Draw;
begin
end;
{------------------------------------------------------}
procedure TCustomElem.SetCanvas(const Value: TCanvas);
begin
  fCanvas := Value;
end;
{------------------------------------------------------}
procedure TCustomElem.SetFill(const Value: TColor);
begin
  fFillStyle.Color := Value;
end;
{------------------------------------------------------}
procedure TCustomElem.SetOutline(const Value: TColor);
begin
  fOutline := Value;
end;
{------------------------------------------------------}
procedure TCustomElem.SetMerge(const Value: boolean);
begin
  fMerge := Value;
end;
{------------------------------------------------------}
procedure TCustomElem.SetRealWidth(const Value: real);
begin
  fRWidth := Value;
end;
{------------------------------------------------------}
procedure TCustomElem.SetPattern(const Value: THpFillType);
begin
  fFillStyle.Pattern := Value;
end;
{------------------------------------------------------}
procedure TCustomElem.SetCanvasParams;
begin
  fCanvas.Pen.Color := fOutline;
  fCanvas.Brush.Color := fFillStyle.Color;
  fCanvas.Pen.Mode := pmCopy;
  case fFillStyle.Pattern of
    ftSolid : fCanvas.Brush.Style := bsSolid;
    ftLines,
    ftCross : begin
                fCanvas.Brush.Color := fFillStyle.Color;
                if fFillStyle.Pattern = ftLines then
                  case fFillStyle.Data of
                    00 : fCanvas.Brush.Style := bsHorizontal;
                    45 : fCanvas.Brush.Style := bsBDiagonal;
                    90 : fCanvas.Brush.Style := bsVertical;
                    135 : fCanvas.Brush.Style := bsFDiagonal;
                  end
                else if fFillStyle.Pattern = ftCross then
                  begin
                    if fFillStyle.Data = 0 then
                      fCanvas.Brush.Style := bsCross
                    else
                      fCanvas.Brush.Style := bsDiagCross;
                  end;
              end;
    ftShade : fCanvas.Brush.Color := ShadeOfColor(fFillStyle.Color,
                                                  fFillStyle.Data);
  end;
end;
{------------------------------------------------------}
procedure TCustomElem.Assign(var Dst: TCustomElem);
begin
  Dst.fType := fType;
  SetLength(Dst.fNodes, SizeOf(fNodes));
  Move(fNodes, Dst.fNodes, SizeOf(fNodes));
  Dst.fBounds := fBounds;
  Dst.fOutline := fOutline;
  Dst.fOutStyle := fOutStyle;
  Dst.fFillStyle := fFillStyle;
  Dst.fCanvas := fCanvas;
  Dst.fCount := fCount;
  Dst.fRWidth := fRWidth;
  Dst.fMerge := fMerge;
  Dst.fParent := fParent;
end;
{------------------------------------------------------}
{ TElementList }
function TElementList.Add(eType: TElemType; Nodes: Pointer) : boolean;
var
  Elem : TCustomElem;
  R    : TRect;
begin
  Result := false;
  try
    case eType of
      etPline : Elem := TPolyline.Create;
      etPgon  : Elem := TPolygon.Create;
      etArc   : Elem := TArc.Create;
      etPie   : Elem := TPie.Create;
      etWin   : Elem := TWindow.Create;
      etFullWin  : Elem := TFullWindow.Create;
      etText  : Exit;
      etImage : Elem := THpImage.Create;
      else Exit;
    end;
  except
    Exit;
  end;
  if eType = etText then
    Elem.AddPoints(PTextInfoType(Nodes))
  else if eType = etImage then
    Elem.AddPoints(TImageDataPtr(Nodes))
  else if eType <> etFullWin then
    Elem.AddPoints(THpNodeList(Nodes));
  if (eType < etWin) and (not Assigned(Elem.fNodes)) then
    begin
      Elem.Free;
      Exit;
    end;
  Inc(fCount);
  if Assigned(fCanvas) then
    Elem.ECanvas := fCanvas
  else
    Elem.ECanvas := nil;
  Elem.Parent := Self;
  if eType < etWin then
    begin
      Elem.fFillStyle := fFill;
      Elem.fOutline := fOutline;
      Elem.fOutStyle := fLineStyle;
      Elem.fRWidth := fWidth;
      Elem.fMerge := fMerge;
    end;
  fList.Add(Elem);
  if eType = etFullWin then
    begin
      fDoBounds := True;
      Result := True;
      Exit;
    end;
  // Calculate list bounds if element is a window
  // otherwise calculate real bounds.
  if (EType = etWin) then //or fDoBounds then
    R := fListBounds
  else
    R := fRealBounds;
  if Elem.Bounds.Left < R.Left then
    R.Left := Elem.Bounds.Left;
  if Elem.Bounds.Right > R.Right then
    R.Right := Elem.Bounds.Right;
  if Elem.Bounds.Top > R.Top then
    R.Top := Elem.Bounds.Top;
  if Elem.Bounds.Bottom < R.Bottom then
    R.Bottom := Elem.Bounds.Bottom;
  fListBounds := R;
  if (EType <> etWin) then
    begin
      fView := R;
      fRealBounds := R;
    end;
  fDoBounds := not (EType = etWin);
  Result := True;
end;
{------------------------------------------------------}
procedure TElementList.Clear;
begin
  fList.Clear;
  SetDefaults;
end;
{------------------------------------------------------}
// Compares two items
function TElementList.Compare(Val1, Val2, Op: integer): integer;
begin
  case Op of
    1 : if Val1 > Val2 then
          Result := Val1
        else
          Result := Val2;
    2 : if Val1 < Val2 then
          Result := Val1
        else
          Result := Val2;
  end;
end;
{------------------------------------------------------}
function TElementList.RealWidth(Wd : real) : integer;
begin
  Result := Round(Wd / 25.4 * fCanvasRes * fXScale);
  if Result = 0 then
    Result := 1;
end;
{------------------------------------------------------}
procedure TElementList.SetDefaults;
begin
  fCount := 0;
  fOutline := clBlack;
  fFill.Color := clWhite;
  fFill.Pattern := ftSolid;
  fFill.Data := 100;
  fLineStyle := psSolid;
  fWidth := 0.35;
  fClip := 0;
  fViewReady := False;
  fViewOk := False;
  fListBounds.Top := -1 * (MaxInt - 1);
  fListBounds.Bottom := MaxInt;
  fListBounds.Left := MaxInt;
  fListBounds.Right := -1 * (MaxInt - 1);
  fRealBounds := fListBounds;
  fView := fListBounds;
  fAborted := False;
  fDoBounds := True;
end;
{------------------------------------------------------}
constructor TElementList.Create;
begin
  fCanvas := nil;
  fList := nil;
  SetDefaults;
  fList := TObjectList.Create;
  fUseTranspar := True;
end;
{------------------------------------------------------}
destructor TElementList.Destroy;
begin
  Clear;
  fList.Free;
  inherited;
end;
{------------------------------------------------------}
function TElementList.DrawList: integer;
var
  I : integer;
  Elem : TCustomElem;
begin
  if fDrawing then
    Exit;
  if fViewReady and fViewOk and Assigned(fCanvas) then
    begin
      fDrawing := True;
      fCanvasRes := GetDeviceCaps(fCanvas.Handle, LOGPIXELSX);
      for I := 0 to fList.Count - 1 do
        begin
          Elem := TCustomElem(fList.Items[I]);
          if Elem.ElementType = etFullWin then
            Elem.Draw
          else if InRange(Elem) then
            try //begin
              if Elem.ElementType <> etImage then
                fCanvas.Pen.Width := RealWidth(Elem.fRWidth);
              Elem.Draw;
            except
              on E : Exception do
                begin
                  MessageDlg(E.Message, mtError, [mbOk], 0);
                  fResources := False;
                  if fClip <> 0 then
                    begin
                      SelectClipRgn(fCanvas.Handle, 0);
                      DeleteObject(fClip);
                    end;
                  Exit;
                end;
            end;
          Application.ProcessMessages;
          if fAborted then
            Break;
        end;
      if fClip <> 0 then
        begin
          SelectClipRgn(fCanvas.Handle, 0);
          DeleteObject(fClip);
        end;
      fDrawing := False;
    end;
end;
{------------------------------------------------------}
function TElementList.GetType(const Index: integer): TElemType;
begin
  if Index >= fCount then
    raise Exception.Create('Node index out of bounds');
  Result := TCustomElem(fList.Items[Index]).ElementType;
end;
{------------------------------------------------------}
// This function checks if given element is at least partially visible.
function TElementList.InRange(Elem: TCustomElem): boolean;
var
  S : boolean;
  Rv, Re : TRect;
begin
  Result := True;
  Rv := fView;
  Re := Elem.Bounds;
  // Normalize view and element rectangles
  if fView.Left > fView.Right then
    begin
      Rv.Left := fView.Right;
      Rv.Right := fView.Left;
    end;
  if fView.Top < fView.Bottom then
    begin
      Rv.Top := fView.Bottom;
      Rv.Bottom := fView.Top;
    end;
  if Elem.Bounds.Left > Elem.Bounds.Right then
    begin
      Re.Left := Elem.Bounds.Right;
      Re.Right := Elem.Bounds.Left;
    end;
  if Elem.Bounds.Top < Elem.Bounds.Bottom then
    begin
      Re.Top := Elem.Bounds.Bottom;
      Re.Bottom := Elem.Bounds.Top;
    end;
  if (Re.Bottom > Rv.Top) or (Re.Top < Rv.Bottom) or (Re.Right < Rv.Left) or
     (Re.Left > Rv.Right) then
    Result := False;
end;
{------------------------------------------------------}
procedure TElementList.SetCanvas(const Value: TCanvas);
var
  I : integer;
begin
  fCanvas := Value;
  if (Assigned(fCanvas)) and (fList.Count > 0) then
    for I := 0 to fList.Count - 1 do
      TCustomElem(fList[I]).fCanvas := Value;
end;
{------------------------------------------------------}
// Calculate world to device transformation parameters.
procedure TElementList.SetShiftAndScale;
var
  S1, S2 : integer;
  X, Y   : real;
begin
  if (fView.Left = fView.Right) or
     (fView.Top = fView.Bottom) or
     (fView.Left = fView.Right) or
     (fView.Top = fView.Bottom) then
    begin
      fViewOk := False;
      Exit;
    end;
  X := (fViewPort.Right - fViewPort.Left) / (fView.Right - fView.Left);
  if X < 0 then
    S1 := -1
  else
    S1 := 1;
  X := Abs(X);
  Y := (fViewPort.Top - fViewPort.Bottom) / (fView.Top - fView.Bottom);
  if Y < 0 then
    S2 := -1
  else
    S2 := 1;
  Y := Abs(Y);
  if X > Y then
    begin
      if Y = 0 then
        Exit;
      fXScale := S1 * Y;
      fYScale := S2 * Y;
      if fListBounds.Left < fListBounds.Right then
        fView.Right := fView.Left + Round((fViewPort.Right - fViewPort.Left) / Abs(fYScale))
      else
        fView.Left := fView.Right + Round((fViewPort.Right - fViewPort.Left) / Abs(fYScale));
    end
  else
    begin
      if X = 0 then
        Exit;
      fXScale := S1 * X;
      fYScale := S2 * X;
      if fListBounds.Top > fListBounds.Bottom then
        fView.Top := fView.Bottom + Round((fViewPort.Bottom - fViewPort.Top) / Abs(fXScale))
      else
        fView.Bottom := fView.Top + Round((fViewPort.Bottom - fViewPort.Top) / Abs(fXScale));
    end;
  fXShift := Round(fViewPort.Left  - fXScale * fView.Left);
  fYShift := Round(fViewPort.Bottom  - fYScale * fView.Bottom);
  fViewOk := True;
end;
{------------------------------------------------------}
procedure TElementList.SetViewFromScreen(const Value: TRect);
var
  I : integer;
begin
  fView.TopLeft := DevToWorld(Value.TopLeft);
  fView.BottomRight := DevToWorld(Value.BottomRight);
  if ((fListBounds.Left < fListBounds.Right) and (fView.Left > fView.Right)) or
     ((fListBounds.Left > fListBounds.Right) and (fView.Left < fView.Right)) then
    begin
      I := fView.Left;
      fView.Left := fView.Right;
      fView.Right := I;
    end;
  if ((fListBounds.Top < fListBounds.Bottom) and (fView.Top > fView.Bottom)) or
     ((fListBounds.Top > fListBounds.Bottom) and (fView.Top < fView.Bottom)) then
    begin
      I := fView.Top;
      fView.Top := fView.Bottom;
      fView.Bottom := I;
    end;
  if fViewReady then
    SetShiftAndScale;
end;
{------------------------------------------------------}
procedure TElementList.SetView(const Value: TRect);
var
  I : integer;
begin
  fView := Value;
  if fViewReady then
    SetShiftAndScale;
end;
{------------------------------------------------------}
procedure TElementList.SetViewPort(const Value : TRect);
begin
  fViewPort := Value;
  if fViewReady then
    SetShiftAndScale;
end;
{------------------------------------------------------}
procedure TElementList.SetViewReady(const Value: boolean);
begin
  fViewReady := Value;
end;
{------------------------------------------------------}
function TElementList.DevToWorld(const DPoint: TPoint): TPoint;
begin
  Result.X := Round((DPoint.X - fXShift) / fXScale);
  Result.Y := Round((DPoint.Y - fYShift) / fYScale);
end;
{------------------------------------------------------}
function TElementList.WorldToDev(const WPoint: TPoint): TPoint;
begin
  Result.X := Round(WPoint.X * fXScale + fXShift);
  Result.Y := Round(WPoint.Y * fYScale + fYShift);
end;
{------------------------------------------------------}
procedure TElementList.InitView(const ViewPort : TRect);
begin
  fViewPort := ViewPort;
  fView := fListBounds;
  SetShiftAndScale;
//  if fViewReady then
//    DrawList;
end;
{------------------------------------------------------}
{ TPolyline }
// Add points and calculate bounding box.
procedure TPolyline.AddPoints(Src: THpNodeList);
var
  I : integer;
begin
  fCount := Src.Count;
  if fCount = 0 then
    Exit;
  try
    SetLength(fNodes, fCount);
  except
    fNodes := nil;
    Exit;
  end;
  for I := 0 to fCount - 1 do
    begin
      fNodes[I] := Src.Get(I);
      if fNodes[I].X < fBounds.Left then
        fBounds.Left := fNodes[I].X;
      if fNodes[I].X > fBounds.Right then
        fBounds.Right := fNodes[I].X;
      if fNodes[I].Y < fBounds.Bottom then
        fBounds.Bottom := fNodes[I].Y;
      if fNodes[I].Y > fBounds.Top then
        fBounds.Top := fNodes[I].Y;
    end;
end;
{------------------------------------------------------}
constructor TPolyline.Create;
begin
  inherited;
  fNodes := nil;
  fType := etPline;
  fCount := 0;
end;
{------------------------------------------------------}
destructor TPolyline.Destroy;
begin
  inherited;
end;
{------------------------------------------------------}
procedure TPolyline.Draw;
begin
  if not Assigned(fCanvas) then
    Exit;
  ScaleNodes;
  if not Assigned(fTempNodes) then
    Exit;
  fCanvas.Brush.Style := bsClear;
  fCanvas.Pen.Color := fOutline;
  fCanvas.Pen.Style := fOutStyle;
  fCanvas.Pen.Mode := pmCopy;
  fCanvas.Polyline(fTempNodes);
  fTempNodes := nil;
//  FreeMem(fTempNodes, fCount * SizeOf(TPoint));
end;
{------------------------------------------------------}
{ TPolygon }
procedure TPolygon.AddPoints(Src: THpNodeList);
var
  I : integer;
begin
  fCount := Src.Count;
  if fCount = 0 then
    Exit;
  try
    SetLength(fNodes, fCount);
  except
    fNodes := nil;
    Exit;
  end;
  for I := 0 to fCount - 1 do
    begin
      fNodes[I] := Src.Get(I);
      if fNodes[I].X < fBounds.Left then
        fBounds.Left := fNodes[I].X;
      if fNodes[I].X > fBounds.Right then
        fBounds.Right := fNodes[I].X;
      if fNodes[I].Y < fBounds.Bottom then
        fBounds.Bottom := fNodes[I].Y;
      if fNodes[I].Y > fBounds.Top then
        fBounds.Top := fNodes[I].Y;
    end;
end;
{------------------------------------------------------}
constructor TPolygon.Create;
begin
  inherited;
  fNodes := nil;
  fType := etPgon;
  fCount := 0;
end;
{------------------------------------------------------}
destructor TPolygon.Destroy;
begin
  inherited;
end;
{------------------------------------------------------}
procedure TPolygon.Draw;
var
  I, X, Y : integer;
  P, P1 : TPoint;
  DoMerge : boolean;
  fBitmap : TBitmap;
begin
  if not Assigned(fCanvas) then
    Exit;
  ScaleNodes;
  if not Assigned(fTempNodes) then
    Exit;
  P := fParent.WorldToDev(Self.fBounds.TopLeft);
  P1 := fParent.WorldToDev(Self.fBounds.BottomRight);
  X := P.X;
  Y := P.Y;
  fBmW := Abs(P1.X - P.X + 1);
  fBmH := Abs(P1.Y - P.Y + 1);
  DoMerge := fMerge and (fBmW > 2) and (fBmH > 2);
//  DoMerge := False;
  if DoMerge or fParent.fUseTranspar then
    begin
      try
        fBitmap := TBitmap.Create;
      except
        ShowMessage('Cannot create bitmap');
        Exit;
      end;
      fBitmap.Width := fBmW;
      fBitmap.Height := fBmH;
      for I := 0 to fCount - 1 do
        begin
          fTempNodes[I].X := fTempNodes[I].X - X;
          fTempNodes[I].Y := fTempNodes[I].Y - Y;
        end;
      fBitmap.Canvas.Pen.Style := psClear;
      fBitmap.Canvas.Brush.Color := fParent.fBkColor;
      fBitmap.Canvas.Rectangle(0, 0, fBmW, fBmH);
      fBitmap.Canvas.Brush.Color := fFillStyle.Color;
      fBitmap.Canvas.Pen.Mode := pmCopy;
      case fFillStyle.Pattern of
        ftSolid : fBitmap.Canvas.Brush.Style := bsSolid;
        ftLines,
        ftCross : begin
                  if fFillStyle.Pattern = ftLines then
                    case fFillStyle.Data of
                      00 : fBitmap.Canvas.Brush.Style := bsHorizontal;
                      45 : fBitmap.Canvas.Brush.Style := bsBDiagonal;
                      90 : fBitmap.Canvas.Brush.Style := bsVertical;
                      135 : fBitmap.Canvas.Brush.Style := bsFDiagonal;
                    end
                  else if fFillStyle.Pattern = ftCross then
                    begin
                      if fFillStyle.Data = 0 then
                          fBitmap.Canvas.Brush.Style := bsCross
                        else
                          fBitmap.Canvas.Brush.Style := bsDiagCross;
                      end;
                  end;
          ftShade : fBitmap.Canvas.Brush.Color := ShadeOfColor(fFillStyle.Color,
                                                        fFillStyle.Data);
      end;
      fBitmap.Canvas.Polygon(fTempNodes);
      DrawTransparent(fBitmap.Handle, X, Y, fParent.fBkColor);
      fBitmap.Free;
    end
  else
    begin
      SetCanvasParams;
      fCanvas.Pen.Style := psClear;
      fCanvas.Polygon(fTempNodes);
    end;
  fTempNodes := nil;
//  FreeMem(fTempNodes, fCount * SizeOf(TPoint));
end;
{------------------------------------------------------}
{ TArc }
procedure TArc.AddPoints(Src: THpNodeList);
begin
  if Src.Count < 4 then
    Exit;
  fCount := 4;
  SetLength(fNodes, 4);
  fBounds.Left := Src.Get(0).X;
  fBounds.Bottom := Src.Get(0).Y;
  fBounds.Right := Src.Get(1).X;
  fBounds.Top := Src.Get(1).Y;
  fNodes[0] := Point(fBounds.Left, fBounds.Bottom);
  fNodes[1] := Point(fBounds.Right, fBounds.Top);
  fNodes[2] := Src.Get(2);
  fNodes[3] := Src.Get(3);
  Src.Clear;
end;
{------------------------------------------------------}
constructor TArc.Create;
begin
  inherited;
  fNodes := nil;
  fType := etArc;
  fCount := 0;
end;
{------------------------------------------------------}
destructor TArc.Destroy;
begin
  inherited;
end;
{------------------------------------------------------}
procedure TArc.Draw;
begin
  if not Assigned(fCanvas) then
    Exit;
  ScaleNodes;
  if not Assigned(fTempNodes) then
    Exit;
  fCanvas.Brush.Style := bsClear;
  fCanvas.Pen.Color := fOutline;
  fCanvas.Pen.Style := fOutStyle;
  fCanvas.Pen.Mode := pmCopy;
  Arc(fCanvas.Handle, fTempNodes[0].X, fTempNodes[0].Y,
       fTempNodes[1].X, fTempNodes[1].Y,
       fTempNodes[2].X, fTempNodes[2].Y,
       fTempNodes[3].X, fTempNodes[3].Y);
//  FreeMem(fTempNodes, fCount * SizeOf(TPoint));
  fTempNodes := nil;
end;
{------------------------------------------------------}
{ TCircle }
procedure TCircle.AddPoints(Src: THpNodeList);
begin
  if Src.Count < 2 then
    Exit;
  fCount := 2;
  fBounds.Left := Src.Get(0).X;
  fBounds.Bottom := Src.Get(0).Y;
  fBounds.Right := Src.Get(1).X;
  fBounds.Top := Src.Get(1).Y;
  SetLength(fNodes, 2);
  fNodes[0] := Point(fBounds.Left, fBounds.Bottom);
  fNodes[1] := Point(fBounds.Right, fBounds.Top);
  Src.Clear;
end;
{------------------------------------------------------}
constructor TCircle.Create;
begin
  inherited;
  fCount := 0;
  fNodes := nil;
  fType := etCircle;
  fCount := 0;
end;
{------------------------------------------------------}
destructor TCircle.Destroy;
begin
  inherited;
end;
{------------------------------------------------------}
procedure TCircle.Draw;
begin
  if not Assigned(fCanvas) then
    Exit;
  ScaleNodes;
  if not Assigned(fTempNodes) then
    Exit;
  fCanvas.Brush.Style := bsClear;
  fCanvas.Pen.Color := fOutline;
  fCanvas.Pen.Style := fOutStyle;
  fCanvas.Pen.Mode := pmCopy;
  Ellipse(fCanvas.Handle, fTempNodes[0].X, fTempNodes[0].Y,
          fTempNodes[1].X, fTempNodes[1].Y);
//  FreeMem(fTempNodes, fCount * SizeOf(TPoint));
  fTempNodes := nil;
end;
{------------------------------------------------------}
{ TWindow }
procedure TWindow.AddPoints(Src: THpNodeList);
begin
  if (not Assigned(Src)) or (Src.Count = 0) then
    fCount := -1
  else if Src.Count = 2 then
    begin
      fCount := 2;
      SetLength(fNodes, 2);
      fBounds.Left := Src.Get(0).X;
      fBounds.Bottom := Src.Get(0).Y;
      fBounds.Right := Src.Get(1).X;
      fBounds.Top := Src.Get(1).Y;
      fNodes[0] := Point(fBounds.Left, fBounds.Bottom);
      fNodes[1] := Point(fBounds.Right, fBounds.Top);
      Src.Clear;
    end;
end;
{------------------------------------------------------}
constructor TWindow.Create;
begin
  inherited;
  fNodes := nil;
  fType := etWin;
  fCount := 0;
end;
{------------------------------------------------------}
destructor TWindow.Destroy;
begin
  inherited;
end;
{------------------------------------------------------}
procedure TWindow.Draw;
var
  H : hRgn;
begin
  ScaleNodes;
  SelectClipRgn(fCanvas.Handle, 0);
  if Assigned(fTempNodes) then
    begin
      H := CreateRectRgn(fTempNodes[0].X, fTempNodes[0].Y,
                         fTempNodes[1].X, fTempNodes[1].Y);
      if H <> 0 then
        begin
          SelectClipRgn(fCanvas.Handle, 0);
          if fParent.ClipArea <> 0 then;
            DeleteObject(fParent.ClipArea);
          SelectObject(fCanvas.Handle, H);
          SelectClipRgn(fCanvas.Handle, H);
          fParent.ClipArea := H;
        end;
//      FreeMem(fTempNodes, fCount * SizeOf(TPoint));
      fTempNodes := nil;
    end;
end;
{------------------------------------------------------}
{ TPie }
procedure TPie.AddPoints(Src: THpNodeList);
begin
  if Src.Count < 4 then
    Exit;
  SetLength(fNodes, 4);
  fBounds.Left := Src.Get(0).X;
  fBounds.Bottom := Src.Get(0).Y;
  fBounds.Right := Src.Get(1).X;
  fBounds.Top := Src.Get(1).Y;
  fNodes[0] := Point(fBounds.Left, fBounds.Bottom);
  fNodes[1] := Point(fBounds.Right, fBounds.Top);
  fNodes[2] := Src.Get(2);
  fNodes[3] := Src.Get(3);
  Src.Clear;
end;
{------------------------------------------------------}
constructor TPie.Create;
begin
  inherited;
  fNodes := nil;
  fType := etPie;
  fCount := 0;
end;
{------------------------------------------------------}
destructor TPie.Destroy;
begin
  inherited;
end;
{------------------------------------------------------}
procedure TPie.Draw;
begin

end;
{------------------------------------------------------}
{ TText }
procedure TText.AddPoints(Src: PTextInfoType);
begin
  SetLength(fText, Length(Src^.Data));
  fText := Copy(Src^.Data, 1, Length(Src^.Data));
  fPoint := Src^.Point;
  fAngle := Src^.Angle;
  fHeight := Src^.Height;
end;
{------------------------------------------------------}
constructor TText.Create;
begin
  inherited;
  fType := etText;
  fText := '';
end;
{------------------------------------------------------}
destructor TText.Destroy;
begin
  fText := '';
  inherited;
end;
{------------------------------------------------------}
procedure TText.Draw;
var
  Fnt, OFnt : hFont;
  Prm       : TLogFont;
  X, Y      : integer;
begin
  with Prm do
    begin
      lfHeight := Round(fHeight * Abs(fParent.YScale));
      lfWidth := 0;
      lfEscapement := fAngle * 10;
      lfOrientation := lfEscapement;
      lfWeight := FW_REGULAR;
      lfItalic := 0;
      lfUnderline := 0;
      lfStrikeOut := 0;
      lfCharSet := DEFAULT_CHARSET;
      lfOutPrecision := OUT_TT_PRECIS;
      lfClipPrecision := CLIP_DEFAULT_PRECIS;
      lfQuality := PROOF_QUALITY;
      lfPitchAndFamily := DEFAULT_PITCH or FF_SWISS;
      lfFaceName := 'Arial';
    end;
  Fnt := CreateFontIndirect(Prm);
  if Fnt <> 0 then
    begin
      SetCanvasParams;
      X := Round(fPoint.X * fParent.XScale) + fParent.XShift;
      Y := Round(fPoint.X * fParent.YScale) + fParent.YShift;
      OFnt := SelectObject(fCanvas.Handle, Fnt);
      fCanvas.TextOut(X, Y, fText);
      SelectObject(fCanvas.Handle, OFnt);
      DeleteObject(Fnt);
    end;
end;
{---------------------------------------------------}
{ TFullWindow }
constructor TFullWindow.Create;
begin
  inherited;
  fCount := 0;
  fType := etFullWin;
end;
{---------------------------------------------------}
destructor TFullWindow.Destroy;
begin
  inherited;
end;
{---------------------------------------------------}
procedure TFullWindow.Draw;
begin
  SelectClipRgn(fCanvas.Handle, 0);
  if fParent.ClipArea <> 0 then
    DeleteObject(fParent.ClipArea);
end;
{---------------------------------------------------}
{ THpImage }
procedure THpImage.AddPoints(Src: TImageDataPtr);
begin
  inherited;
  fCount := Src^.Bytes;
  try
    GetMem(fImage, fCount);
//    SetLength(fImage, fCount);
  except
    fImage := nil;
    fCount := 0;
    Exit;
  end;
  Move(Src^.Image^[0], fImage^[0], fCount);
  fDPI := Src^.Resolut;
  fWidth := Src^.Width;
  fHeight := Src^.Height;
  fBounds.Left := Src^.X;
  fBounds.Bottom := Src^.Y;
  fBounds.Right := fBounds.Left + Round(Src^.Width * HPGLSteps / fDPI);
  fBounds.Top := fBounds.Bottom + Round(Src^.Height * HPGLSteps / fDPI);
end;
{---------------------------------------------------}
constructor THpImage.Create;
begin
  inherited;
end;
{---------------------------------------------------}
destructor THpImage.Destroy;
begin
  fImage := nil;
  inherited;
end;
{---------------------------------------------------}
procedure THpImage.Draw;
var
  P1     : TPoint;
  dW, dH : integer;
  hBmp   : HBITMAP;
  mIDC,
  hIDC   : HDC;
  BClr   : TColor;
begin
  if (not Assigned(fImage)) or (fCount = 0) then
    Exit;
  dW := Abs(Round((fBounds.Right - fBounds.Left) * fParent.XScale));
  dH := Abs(Round((fBounds.Top - fBounds.Bottom) * fParent.YScale));
  P1 := Parent.WorldToDev(Point(fBounds.Left, fBounds.Top));
  hBmp := CreateBitmap(fWidth, fHeight, 1, 1, fImage);
  if hBmp > 0 then
    begin
      mIDC := GetDC(fCanvas.Handle);
      hIDC := CreateCompatibleDC(mIDC);
      if hIDC = 0 then
        begin
          DeleteObject(hBmp);
          Exit;
        end;
      SelectObject(hIDC, hBmp);
      BClr := SetBkColor(fCanvas.Handle, fParent.fBkColor); //clWhite);
      StretchBlt(fCanvas.Handle, P1.X, P1.Y, dW, dH,
                 hIDC, 0, 0, fWidth, fHeight, SRCAND);
      DeleteDC(hIDC);
      ReleaseDC(fCanvas.Handle, mIDC);
      DeleteObject(hBmp);
      SetBkColor(fCanvas.Handle, BClr);
    end;
end;
{---------------------------------------------------}
end.
